"""All delatore outputs"""
from threading import Event

from apubsub import Service

from .telegram import BotRunner
from ..configuration import InstanceConfig


async def start_outputs(service: Service, stop_event: Event, config: InstanceConfig):
    await BotRunner(service, stop_event, config).start()



"""Base Source implementation"""

import asyncio
import json
import logging
from abc import ABC, ABCMeta, abstractmethod
from inspect import isabstract
from typing import Optional, Union

from apubsub.client import Client

from ..configuration import DEFAULT_INSTANCE_CONFIG, InstanceConfig, SOURCES_CFG
from ..configuration.static import SourceConfiguration

Json = Union[dict, list]

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


class NoUpdates(Exception):
    """Raised when source has no updates"""


class SourceMeta(ABCMeta):
    """Source class meta setting up data from configuration"""

    CONFIG_ID: str
    TOPIC: str
    config: SourceConfiguration

    def __init__(cls, name, bases=(), dct=None):  # pylint:disable=redefined-builtin
        super().__init__(name, bases, dct)
        if not isabstract(cls):  # skip configuration for abstracts
            assert hasattr(cls, 'CONFIG_ID')
            cls.config = SOURCES_CFG[cls.CONFIG_ID]
            cls.TOPIC = cls.config.publishes


class Source(ABC, metaclass=SourceMeta):
    """Source API posting updates to """

    CONFIG_ID: str = NotImplemented

    client: Client
    config: SourceConfiguration
    instance_config: InstanceConfig

    def __init__(self, client: Client,
                 ignore_duplicates=True,
                 instance_config: InstanceConfig = DEFAULT_INSTANCE_CONFIG):
        self.client = client
        self.polling_interval = self.config.timings.polling_interval
        self.request_timeout = self.config.timings.request_timeout
        self.ignore_duplicates = ignore_duplicates
        self.instance_config = instance_config

    @abstractmethod
    async def get_update(self) -> Optional[dict]:
        """Get source update"""

    async def start(self, stop_event: asyncio.Event):
        """Start processing updates"""
        last = None
        name = type(self).__name__
        LOGGER.info('Source %s started', name)
        while not stop_event.is_set():
            try:
                new = await asyncio.wait_for(self.get_update(), self.request_timeout)
            except (asyncio.TimeoutError, NoUpdates):
                continue
            if _same_status(new, last) and self.ignore_duplicates:
                continue
            LOGGER.debug('New data received from source: %s\ndata:\n%s', name, new)
            json_message = json.dumps(new)
            await self.client.publish(self.TOPIC, json_message)
            last = new
            LOGGER.debug('Wait for new data for %s', self.polling_interval)
            await asyncio.sleep(self.polling_interval)


def _same_status(new: dict, last: dict) -> bool:
    if last is None:
        return False
    if new == last:
        return True
    this_statuses = {st['name']: st['status'] for st in new['status_list']}
    another_statuses = {st['name']: st['status'] for st in last['status_list']}
    return this_statuses == another_statuses



"""Configurations fixed for library"""
from dataclasses import dataclass
from typing import Dict, List

import tyaml
from ocomone import Resources


@dataclass(frozen=True)
class Configuration:
    """Base configuration class"""
    name: str
    params: dict


@dataclass(frozen=True)
class OutputConfiguration(Configuration):
    """Single output service configuration"""
    subscriptions: List[str]


@dataclass(frozen=True)
class Timings:
    polling_interval: float
    request_timeout: float


@dataclass(frozen=True)
class SourceConfiguration(Configuration):
    """Single source service configuration"""
    publishes: str
    timings: Timings


_CONFIGS = Resources(__file__)


def _cfg_load(cfg_file: str, cfg_class):
    with open(cfg_file, 'r') as src_cfg:
        configs = tyaml.load(src_cfg, cfg_class)  # type: List[Configuration]
    result = {cfg.name: cfg for cfg in configs}
    return result


SOURCES_CFG: Dict[str, SourceConfiguration] = _cfg_load(_CONFIGS['sources.yaml'], List[SourceConfiguration])
OUTPUTS_CFG: Dict[str, OutputConfiguration] = _cfg_load(_CONFIGS['outputs.yaml'], List[OutputConfiguration])



"""Parsing command arguments """
import logging
import shlex

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)

__all__ = ['parse_command', 'CommandParsingError']


class CommandParsingError(Exception):
    """Error during command parsing"""


def parse_command(message):
    """Parsing command arguments to arguments list"""
    LOGGER.debug('Got message: %s', message)
    try:
        cmd, target, *args = shlex.split(message)
    except ValueError:
        raise CommandParsingError('No target provided for `/status` command')
    depth = None
    detailed = args.pop(0) if args else None
    try:
        depth = int(args.pop(0))
    except IndexError:
        depth = 1
    except ValueError:
        raise CommandParsingError(f'Invalid depth value: {depth}')
    if args:
        raise CommandParsingError(f'Command {cmd} got unexpected arguments: {args}')
    return target, detailed, depth



"""Conversion of source data to markdown v2"""
from aiogram.utils.markdown import escape_md, link

from ...outputs.telegram.emoji import EMOJI
from ...unified_json import Status


def _status_to_md_row(status_record):
    status_emoji = replace_emoji(status_record['status'])
    name = escape_md(status_record['name'])
    timestamp = escape_md(status_record['timestamp'])
    url = status_record['details_url']
    if url:
        timestamp = link(timestamp, url)
    else:
        timestamp = f'`{timestamp}`'
    details = rf'{name} \({timestamp}\)'
    return rf'{status_emoji}  —  {details}'


def convert(data: dict):
    """Convert dict or list to nice-looking telegram markdown"""
    source = escape_md(data['source'])
    src_row = f'*From {source}*'
    if 'status_list' in data:
        status_list = '\n'.join(
            _status_to_md_row(status) for status in data['status_list']
        )
    else:
        status_list = f'Error: {data["error"]}'
    return f'{src_row}\n{status_list}'


def replace_emoji(status):
    return EMOJI.get(Status(status), '')



"""Library for monitoring CSM scripts"""

import logging
import os
from logging.handlers import TimedRotatingFileHandler

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


def setup_logger(logger: logging.Logger, log_name: str = None,
                 log_format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                 log_dir='./logs'):
    """Setup logger with ``INFO`` level stream output + ``DEBUG`` level rotating file log

    :param logger: logger to be wrapped. Do not wrap same logger twice!
    :param log_name: name of logfile to be created
    :param log_format: format of log output
    :param log_dir: path to rotating file log directory
    """
    formatter = logging.Formatter(log_format)
    if log_name is None:
        log_name = logger.name.lower()
    try:
        # debug+ messages goes to log file
        os.makedirs(log_dir, exist_ok=True)
        f_hdl = TimedRotatingFileHandler(f'{log_dir}/{log_name}.log', encoding='utf-8', backupCount=10, when='midnight',
                                         utc=True)
        f_hdl.setLevel(logging.DEBUG)
        f_hdl.setFormatter(formatter)
        logger.addHandler(f_hdl)
    except OSError:
        pass
    # info+ messages goes to stream
    s_hdl = logging.StreamHandler()
    s_hdl.setLevel(logging.INFO)
    s_hdl.setFormatter(formatter)
    logger.addHandler(s_hdl)


setup_logger(LOGGER, 'delatore', log_dir=os.path.expanduser('~/.delatore/log'))



"""Configuration loaded from external source"""
import configparser
import os
from typing import NamedTuple, Optional

from ocomone import Resources

RESOURCES = Resources('.', './config')


class InstanceConfig(NamedTuple):
    """Bot configuration container"""
    token: str
    chat_id: str
    influx_password: str
    awx_auth_token: str
    proxy: Optional[str] = None


def read_config(config_file=RESOURCES['config.ini']):
    """Read configuration from configuration file"""
    config = configparser.ConfigParser()
    config.read(config_file)
    defaults = config.defaults()
    if defaults:
        token = defaults.get('token', os.getenv('token'))
        chat_id = defaults.get('chat_id', os.getenv('chat_id'))
        influx_password = defaults.get('influx_password', os.getenv('INFLUX_PASSWORD'))
        awx_auth_token = defaults.get('awx_auth_token', os.getenv('AWX_AUTH_TOKEN'))
        proxy = defaults.get('proxy', None)
        return InstanceConfig(token, chat_id, influx_password, awx_auth_token, proxy)
    return InstanceConfig(
        os.getenv('token'),
        os.getenv('chat_id'),
        os.getenv('INFLUX_PASSWORD'),
        os.getenv('AWX_AUTH_TOKEN')
    )


DEFAULT_INSTANCE_CONFIG = read_config()



import re
from typing import Dict

from ...unified_json import Status

EMOJI = {
    Status.FAIL: '❌',
    Status.RUNNING: '🏃',
    Status.OK: '✅',
    Status.CANCELED: '⛔',
    Status.NO_DATA: '❔',
}


def replace_emoji(source: str, alias_map: Dict[str, str], replacement='%e %a'):
    """Replace all aliases in text with given emoji

    `alias_map` is map of `{"alias": "<emoji>"}`, where "alias" is replaced according to `replacement` schema

    `replacement` has following syntax: `%e` for emoji, `%a` for alias, others characters is treated as plain text
    """
    pattern = re.compile('|'.join(alias_map.keys()))

    def _replace_re(match):
        alias = match.group(0)
        rep_pattern = replacement. \
            replace('%e', alias_map[alias]). \
            replace('%a', alias)
        return rep_pattern
    text = pattern.sub(_replace_re, source)
    return text
